!***********************************************************************
 module atoms
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************

 implicit none

! set all entities only to be accessible within this module by default
 private

! declarations available externally
 public :: atom_type

! procedures available externally
 public :: create_atoms, destroy_atoms, get_atoms, put_atoms, rd_atoms_input, &
          & rd_atom_state, store_atom_state, zero_momenta_a

 type atom_type
  double precision, dimension(:,:), allocatable :: r,p,f
  double precision, dimension(:), allocatable :: eint
  integer :: n
 end type atom_type

 contains
!***********************************************************************
 subroutine create_atoms(a, d)
!***********************************************************************
 type (atom_type), intent(inout) :: a
 integer :: error, n, d

 n = a%n

! allocate arrays for atoms
  allocate(a%r(n,d), a%p(n,d),a%f(n,d),a%eint(n), stat=error)
  if(error .ne. 0) stop 'atoms.create_atoms: allocation error'

  end subroutine create_atoms
!***********************************************************************
  subroutine destroy_atoms(a)
!***********************************************************************
  type (atom_type) :: a

! deallocate arrays for atoms
   if(allocated(a%r)) deallocate(a%r)
   if(allocated(a%p)) deallocate(a%p)
   if(allocated(a%f)) deallocate(a%f)
   if(allocated(a%eint)) deallocate(a%eint)

  end subroutine destroy_atoms
!***********************************************************************
  subroutine get_atoms(a,n,r,p,f,eint)
! unpack the components of the derived type
!***********************************************************************
   type (atom_type), intent(in) :: a
   integer, intent(out), optional :: n
   double precision, dimension(:,:), intent(out), optional :: r,p,f
   double precision, dimension(:), intent(out), optional :: eint

   if(present(n)) n = size(a%r,1)
   if(present(r)) r = a%r
   if(present(p)) p = a%p
   if(present(f)) f = a%f
   if(present(f)) eint = a%eint
   
  end subroutine get_atoms
!***********************************************************************
  subroutine put_atoms(a,n,r,p,f,eint)
! pack the components of the derived type
!***********************************************************************
   type (atom_type), intent(out) :: a
   integer, intent(in), optional :: n
   double precision, dimension(:,:), intent(in), optional :: r,p,f
   double precision, dimension(:), intent(in), optional :: eint

   if(present(n)) a%n = n
   if(present(r)) a%r = r
   if(present(p)) a%p = p
   if(present(f)) a%f = f
   if(present(f)) a%eint = eint

  end subroutine put_atoms
!***********************************************************************
  subroutine store_atom_state(a,sunit)
!***********************************************************************  
  type(atom_type) :: a
  integer :: sunit

!   open(unit=sunit,file=a%fname,&
!  &    form='unformatted',status='unknown',position='rewind')
   write(sunit) a%n,a%r, a%p
!   close(sunit)

  end subroutine store_atom_state
!***********************************************************************
  subroutine rd_atoms_input(atd,header,ifile,iunit)
!***********************************************************************
  type (atom_type) :: atd
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring
  character(len=32) :: ifile      !input file name

  rdstring = ''
  i = 0
  open(iunit,file=ifile,status='old',form='formatted', &
     & position='rewind')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) atd%n
   close(iunit)
  else
   close(iunit)
   stop 'Atoms input data not found - stop'
  endif

  end subroutine rd_atoms_input
!***********************************************************************
  subroutine rd_atom_state(a,iunit)
!***********************************************************************
  type (atom_type) :: a
  integer, intent(in) :: iunit
  integer :: n=0

  read(iunit) n,a%r,a%p
  if(n .ne. a%n) stop 'wrong number of atoms in state file'

  end subroutine rd_atom_state
!***********************************************************************
  subroutine wr_atom(atd,ofile,ounit)
! writes input data to the output file
!***********************************************************************
  type (atom_type), intent(in) :: atd
  integer :: ounit
  character(len=32) :: ofile      !output file name

  open(ounit,file=ofile,status='old',form='formatted',position='append')
  write (ounit,*) 'Initial atom data values'
  write (ounit,*)
  write (ounit,*) 'Input number of atoms   = ', atd%n
  write (ounit,*)
  close(ounit)
  
  end subroutine wr_atom
!***********************************************************************
  subroutine zero_momenta_a(p,dim)
!***********************************************************************
  double precision, dimension(:,:), intent(inout) :: p
  integer :: i, n
  integer, optional :: dim
  
  n = size(p,1)

  if(present(dim)) then
! set total momentum to zero in chosen direction
   p(:,dim) = p(:,dim) - sum(p(:,dim))/n
  else
! set total momentum to zero in each direction
   do i = 1, size(p,2)
    p(:,i) = p(:,i) - sum(p(:,i))/n
   enddo
  endif
    
  end subroutine zero_momenta_a
!***********************************************************************
  end module atoms
