!***********************************************************************
  module barostat
!***********************************************************************
! Pressure control functions.
!
! Values of constants:
! barotype: 0 no thermostat
!           2 Nose-Hoover molecular barostat
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  implicit none

  private

! Public procedures
  public :: rdinput_barostat

  public :: barostat_data_type

  type barostat_data_type
   integer :: barotype  ! barostat type
   double precision :: pin ! input pressure
   double precision :: edot  ! Nose-Hoover barostat multiplier
   double precision :: qpr  ! Nose-Hoover barostat squared relaxation time
   double precision :: volume  ! volume for const pressure
  end type barostat_data_type

  contains
!***********************************************************************
  subroutine rdinput_barostat(prd,iunit)
!***********************************************************************
! Reads barostat variables from input file.
! 
!***********************************************************************
  implicit none
  
  type (barostat_data_type) :: prd
  
  integer :: iunit
  
  read(iunit,*) prd%barotype
  read(iunit,*) prd%pin
  read(iunit,*) prd%qpr
  
  if(.not. (prd%barotype .eq. 0 .or. prd%barotype .eq. 2)) &
  & stop 'barostat:rdinput_barostat invalid value of barotype'

  end subroutine rdinput_barostat
!***********************************************************************
  end module barostat
