!***********************************************************************
  module ensemble_av
!***********************************************************************
! Ensemble averaging using Fortran2008 coarrays.
! Currently, this only works with the Intel Fortran compiler on Linux.
! Unfortunately, it cannot be used with other compilers or Mac OSX!!!
!
! This version: ensemble_av_nocoarray
! Coarray functionality removed to allow running without ensemble
! averaging.
!
!23456789012345678901234567890123456789012345678901234567890123456789012
!***********************************************************************
  use run_control

  implicit none

  private

! Public procedures
  public :: set_frac, init_ensav, create_coarray, create_coarray1, &
            co_mean, co_mean1, co_stderr

! Public data types
  public :: eav_data

! Module data type
  type eav_data
   double precision, dimension(:), allocatable :: codata
   double precision, allocatable :: codata1
   integer :: ni  ! number of images
   integer :: im  ! this image
   character(len=2) :: imlabel  ! image label
  end type eav_data

  contains
!***********************************************************************
  subroutine co_mean(ead,means)
!***********************************************************************
  implicit none
  type(eav_data) :: ead
  double precision, dimension(size(ead%codata)), intent(out) :: means
  integer :: i

  means = 0.d0

  do i = 1, size(ead%codata)
   means(i) = means(i) + ead%codata(i)
  enddo

  means = means/dble(ead%ni)

  end subroutine co_mean
!***********************************************************************
  subroutine co_stderr(ead,means,stderr)
!***********************************************************************
  implicit none
  type(eav_data) :: ead
  double precision, dimension(size(ead%codata)) :: means, stderr
  integer :: i

  stderr = 0.d0

  do i = 1, size(ead%codata)
    stderr(i) = 0.d0
  enddo

  end subroutine co_stderr
!***********************************************************************
  subroutine co_mean1(ead,mean)
!***********************************************************************
  implicit none
  type(eav_data) :: ead
  double precision :: mean
  integer :: i

  mean = 0.d0

  do i = 1, ead%ni
    mean = mean + ead%codata1
  enddo

  mean = mean/dble(ead%ni)

  end subroutine co_mean1
!***********************************************************************
  subroutine create_coarray(ead,a)
! Calculate the ensemble average of an array of properties.
!***********************************************************************
  implicit none
  type(eav_data) :: ead
  double precision, dimension(:) :: a
  integer :: error

! Move data to temporary storage area for ensemble averaging.
  if(allocated(ead%codata)) deallocate(ead%codata)
  allocate(ead%codata(size(a)), stat=error)
  if(error .ne. 0) stop 'allocation error in create_coarray'
  ead%codata = a

  end subroutine create_coarray
!***********************************************************************
  subroutine create_coarray1(ead,a)
! Move data to temporary storage area for ensemble averaging.
!***********************************************************************
  implicit none
  type(eav_data) :: ead
  double precision :: a

  ead%codata1 = a

  end subroutine create_coarray1
!***********************************************************************
  function set_frac(ead)
! Set a unique double precision value between 0 and 1 for this image.
! This is used to seed the random number generator when initializing
! momenta.
!***********************************************************************
  implicit none
  type(eav_data) :: ead
  real :: set_frac

  set_frac = real(ead%im)/real(ead%ni)

  end function set_frac
!***********************************************************************
  subroutine init_ensav(ead, ne)
! Initializes number of images and image id numbers to set up ensemble
! averaging.
!***********************************************************************
  implicit none
  type(eav_data) :: ead
  integer :: ne

  ead%im = 1
  ead%ni = 1

  if(ne .eq. 1) then
   ead%imlabel = ''
  elseif(ne .gt. 1) then
   if(ne .ne. ead%ni) then
    stop 'init_ensav: num_ens_mem does not match num_images'
   else
    write(ead%imlabel,'(a1,i2.2)') '_', ead%im
   endif
  else
   stop 'init_ensav: invalid value of num_ens_mem'
  endif  

  end subroutine init_ensav
!***********************************************************************
  end module ensemble_av
!***********************************************************************
