!***********************************************************************
  subroutine mdstep(rc,box,atd,ded,pf,nl,p,sd,md,ca,spd, &
   da,bs,bb,sys,ewd)
!***********************************************************************
  use run_control
  use integrator
  use gear5
  use simulation_box
  use atoms
  use pair_force
  use neighbour_list
  use properties
  use thermostat
  use sllod
  use molecules
  use constraints
  use species
  use dihedral_force
  use bond_stretch
  use bond_bend
  use system
  use ewald
  use v_sub_library

  implicit none

  type(run_control_type) :: rc
  type(box_data_type) :: box
  type(atom_type) :: atd
  type(integrator_data_type) :: ded
  type(properties_type) :: p
  type(sllod_data_type) :: sd
  type(constraint_set) :: ca
  type(mol_set) :: md
  type(species_set) :: spd
  type(pf_data_type), dimension(spd%npsp) :: pf
  type(nlist_data_type), dimension(spd%npsp) :: nl
  type(dihed_set) :: da
  type(bondstr_set) :: bs
  type(bondbend_set) :: bb
  type(system_data_type) :: sys
  type(ewald_data_type) :: ewd

  double precision, dimension(size(atd%r,1),size(atd%r,2)) :: dr, reqm, peqm
  double precision, dimension(md%nm,size(md%rcm,2)) :: rsh
  double precision, dimension(3,3) :: ptm, pta
  double precision, dimension(3) :: jqm
  double precision :: alpha, zeqm, eeqm, veqm
  integer :: i, j, k

  call init_mdstep

  select case(ded%destype)
  case(1)
   call g5step ! do a gear5 predictor corrector step
  case(2)
   call rk4step(size(atd%r,1), size(atd%r,2)) ! do a RK4 step
  end select

! Zero momenta if required
  do i = 1, 3
   if(rc%kpz(i) .gt. 0 .and. mod(rc%nstep,rc%kpz(i)) .eq. 0) &
    call zero_momenta_m(atd%p, md, spd, i)
  enddo

! Zero y-momentum every step if doing pef
  if(sd%flow_type .eq. 2) call zero_momenta_m(atd%p, md, spd, 2)

! calculate end-step properties
  call calc_props_endstep

! do pbcs - apply pbcs to the cm of each molecule
  call do_pbcs

! write time series
  if(p%do_tseries_now) call wr_property_tseries(p,rc%nstep*ded%tstep,2)

! write snapshot
  if(rc%do_snapshot_now) &
   call wr_snapshot(atd%r, box%boxvec, spd%label(spd%spid), rc, 3)
  
  return

  contains
!***********************************************************************
! internal subroutines for mdstep
!***********************************************************************
  subroutine calc_constraint_forces
!***********************************************************************
! calculate bond constraint forces, add to atom force and calculate
! contribution to atomic pressure tensor
  call add_constraint_forces(atd,md,spd,ca,reqm,peqm,pta)
  if(ca%csolvertype .ne. 0) p%inst(p%i(40)) = ca%cset(1)%its

! calculate thermostat multiplier and add thermostat force to
! equations of motion
   if(sys%td%thermotype .eq. 1) then
! molecular thermostat
! calculate molecular thermostat multiplier
    alpha = gauss_therm_mult(md%pcm, md%fm, sd%flow_type, sd%strain_rate, md%massinv)
! add barostat term to thermostat multiplier if const pressure
    if(sys%prd%barotype .eq. 2) alpha = alpha - sys%prd%edot
    do i = 1, atd%n
     peqm(i,:) = peqm(i,:) - alpha*md%pcm(md%mid(i),:)*spd%mass(spd%spid(i))*md%massinv(md%mid(i))
    enddo
   elseif(sys%td%thermotype .eq. 2) then ! NH kinetic thermostat
    do i = 1, atd%n
     peqm(i,:) = peqm(i,:) - sys%td%zeta*md%pcm(md%mid(i),:)*spd%mass(spd%spid(i))*md%massinv(md%mid(i))
    enddo
   endif
   p%inst(p%i(16)) = alpha

  end subroutine calc_constraint_forces
!***********************************************************************
  subroutine calc_eqm
!***********************************************************************
! add sllod term to rdot equation
  if(sd%flow_type .ne. 0) &
   call add_r_grad_u(reqm, md%rcm(md%mid,:), sd)

! add barostat term to rdot and pdot equations
  if(sys%prd%barotype .ne. 0) then
    do i = 1, atd%n
     reqm(i,:) = reqm(i,:) + sys%prd%edot*md%rcm(md%mid(i),:)
     peqm(i,:) = peqm(i,:) - sys%prd%edot*md%pcm(md%mid(i),:) &
      *spd%mass(spd%spid(i))*md%massinv(md%mid(i))
    enddo
  endif

! equations of motion for NH type thermostats
  if(sys%td%thermotype .eq. 2) then
   zeqm = (p%inst(p%i(20)) - sys%td%tin)/(sys%td%tin*sys%td%tnh)  ! zetadot equation - kinetic
  endif

! equations of motion for barostat
  if(sys%prd%barotype .eq. 2) then
   eeqm = (p%inst(p%i(21)) - sys%prd%pin)*sys%prd%volume/ &
    (md%nm*p%inst(p%i(20))*sys%prd%qpr)  ! edot equation
  elseif(sys%prd%barotype .eq. 3) then
   sys%prd%edot = (p%inst(p%i(21)) - sys%prd%pin)*sys%prd%volume/ &
    (md%nm*p%inst(p%i(20))*sys%prd%qpr)
  endif

  if(sys%prd%barotype .ne. 0) veqm = 3.d0*sys%prd%edot*sys%prd%volume  ! vdot equation
  
  if(sys%td%thermotype .eq. 2) then
   do i = 1, atd%n
    peqm(i,:) = peqm(i,:) - sys%td%zeta*md%pcm(md%mid(i),:)*spd%mass(spd%spid(i))*md%massinv(md%mid(i))
   enddo
  endif

! add sllod term to pdot equation
  if(sd%flow_type .ne. 0) &
   call add_p_grad_u(peqm,md%pcm(md%mid,:) &
    *spread(spd%mass(spd%spid)*md%massinv(md%mid),dim=2, &
    ncopies=size(atd%r,2)), sd)

  end subroutine calc_eqm
!***********************************************************************
  subroutine calc_props_2
!***********************************************************************
! calculate mid-step properties
  if(p%accumulate_properties_now) call calc_props_midstep

! sum internal energy of each molecule
  call calc_eint_mol(md, atd)

  if(p%accumulate_properties_now .or. sys%prd%barotype .ne. 0) call calc_fluxes

! compute kinetic temperatures

! atomic temperature
   p%inst(p%i(10)) = kinetic_temperature(atd%p, &
   & sys%td%ndfa, 1.d0/spd%mass(spd%spid))
! molecular temperature
    p%inst(p%i(20)) = kinetic_temperature(md%pcm, &
    & sys%td%ndfm, md%massinv)
! Apply temperature feedback to pdot equation - for homogeneous system only
    if(sys%td%thermotype .eq. 1 .and. sys%td%tfb .ge. 1.e-9) then
     do j = 1, atd%n
      peqm(j,:) = peqm(j,:) - sys%td%tfb*(p%inst(p%i(20)) - sys%td%tin) &
                & *md%pcm(md%mid(j),:)*md%massinv(md%mid(j))
     enddo
    endif

! squared internal energy/particle
   p%inst(p%i(2)) = p%inst(p%i(1))*p%inst(p%i(1))
! Apply temperature feedback to pdot equation - for homogeneous system only
   if(sys%td%thermotype .eq. 1 .and. sys%td%tfb .ge. 1.e-9) then
    do i = 1, atd%n
     peqm(i,:) = peqm(i,:) - sys%td%tfb*(p%inst(p%i(20)) - sys%td%tin) & 
               & *md%pcm(md%mid(i),:)*md%massinv(md%mid(i))
    enddo
   endif

! compute centre of mass
  do i = 1, 3
   p%inst(p%i(43)-1+i) = sum(atd%r(:,i)*spd%mass(spd%spid))/sum(spd%mass(spd%spid))
  enddo

! compute atomic pressure tensor
   pta = pta + sum_psq_on_m(atd%p, 1.d0/spd%mass(spd%spid))
   do i = 1, spd%npsp
    if(pf(i)%pftype .ne. 0 .and. nl(i)%inab .ne. 0) then
     pta = pta - sum_rij_fij(pf(i)%rij(1:nl(i)%inab,:), pf(i)%fij(1:nl(i)%inab,:))
    endif
   enddo

   pta = pta/box%volume

   do i = 1, 3
    do j = 1, 3
     p%inst(p%i(12)-1+(i-1)*3+j) = pta(i,j)
    enddo
   enddo

   p%inst(p%i(11)) = tr_div_dim(pta)

  end subroutine calc_props_2
!***********************************************************************
  subroutine calc_props_midstep
!***********************************************************************

! add atomic kinetic energy to internal energy
   atd%eint = atd%eint + 0.5d0*sum(atd%p**2,2)/spd%mass(spd%spid)

! atomic pair potential energies
   do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0) then
    if(nl(i)%inab .ne. 0) &
    & call sum_pair_energies(nl(i), pf(i)%uij(1:nl(i)%inab), atd%eint)
    endif
   enddo

! total potential energy for each pair of species
   do i = 1, spd%npsp
    if(pf(i)%pftype .ne. 0) then
     if(nl(i)%inab .ne. 0) p%inst(p%i(6)-1+i) = sum(pf(i)%uij(1:nl(i)%inab))
    endif
   enddo

! short range potential energy (including real space electrostatic)
   p%inst(p%i(3)) = sum(p%inst(p%i(6):p%i(6)-1+spd%npsp))

! short ranged electrostatic energy
  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
    p%inst(p%i(51)) = p%inst(p%i(51)) + pf(i)%esd%ues
    p%inst(p%i(50)) = p%inst(p%i(50)) + pf(i)%esd%pes/box%volume
   elseif(pf(i)%pftype .eq. 5 .or. pf(i)%pftype .eq. 6) then
    p%inst(p%i(51)) = p%inst(p%i(51)) + pf(i)%dsfd%ues
    p%inst(p%i(50)) = p%inst(p%i(50)) + pf(i)%dsfd%pes/box%volume
   endif
  enddo
  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 5 .or. pf(i)%pftype .eq. 6) then
    p%inst(p%i(51)) = p%inst(p%i(51))/(3.d0*box%volume)
    exit
   endif
  enddo

! Fourier, self and self-intramolecular contributions to electrostatic energy
  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
    p%inst(p%i(51)) = p%inst(p%i(51)) + p%inst(p%i(47)) + p%inst(p%i(48)) &
    & + p%inst(p%i(49))
    p%inst(p%i(51)) = p%inst(p%i(51))/(3.d0*box%volume)
    exit
   endif
  enddo

! kinetic energy
   do i = 1, spd%nsp
    if(spd%has_mass(i)) &
     & p%inst(p%i(5)-1+i) = 0.5d0*sum(atd%p(spd%sp(i)%aid,:)**2)/spd%mass(i)
   enddo

! kinetic energy
   p%inst(p%i(4)) = sum(p%inst(p%i(5):p%i(5)-1+spd%nsp))

! internal energy/particle
   p%inst(p%i(1)) = p%inst(p%i(3)) + p%inst(p%i(4))
   if(bs%bstype .ne. 0) p%inst(p%i(1)) = p%inst(p%i(1)) + p%inst(p%i(37))
   if(bb%bbtype .ne. 0) p%inst(p%i(1)) = p%inst(p%i(1)) + p%inst(p%i(57))
   if(da%dfarch .ne. 0) p%inst(p%i(1)) = p%inst(p%i(1)) + p%inst(p%i(41))
   do i = 1, spd%npsp
    if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
     p%inst(p%i(1)) = p%inst(p%i(1)) + p%inst(p%i(47))
     p%inst(p%i(1)) = p%inst(p%i(1)) + p%inst(p%i(48))
     p%inst(p%i(1)) = p%inst(p%i(1)) + p%inst(p%i(49))
     exit
    endif
   enddo

  end subroutine calc_props_midstep
!***********************************************************************
  subroutine calc_props_endstep
!***********************************************************************

! compute total momentum, molecular cm momentum and molecular force
  do i = 1, 3
   p%inst(p%i(42)-1+i) = sum(atd%p(:,i))
   p%inst(p%i(44)-1+i) = sum(md%pcm(:,i))
   p%inst(p%i(45)-1+i) = sum(md%fm(:,i))
  enddo

  p%inst(p%i(7)) = box%volume

   p%inst(p%i(17)) = sd%strain_rate
   p%inst(p%i(18)) = sd%strain
   p%inst(p%i(19)) = sd%strain_per_period

  do i = 1, 3
   do j = 1, 3
    p%inst(p%i(36)-1+(i-1)*3+j) = box%boxvec(i,j)
   enddo
  enddo

  if(ca%csolvertype .ne. 0) then
   do i = 1, md%nf
    p%inst(p%i(39)-1+i) = ca%cset(i)%rmsberr
    p%inst(p%i(52)-1+i) = ca%cset(i)%rmsverr
    p%inst(p%i(40)-1+i) = ca%cset(i)%its
   enddo
  endif

  end subroutine calc_props_endstep
!***********************************************************************
  subroutine calc_pe_forces
!***********************************************************************
! Calculate molecular centre of mass momenta
  call calc_pcm(md)

  call rescale_cm_momenta

! Calculate molecular centre of mass positions
  call calc_rcm(md, spd)

! reform neighbour list if necessary
  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0) then
    if(reform_nlist_now(nl(i))) call form_nlist(nl(i), box, atd%r(nl(i)%aid,:), md)
   endif
  enddo

! calculate pair forces for all interacting pairs of species
  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0 .and. nl(i)%nab .ne. 0) &
    call calculate_pf(nl(i), atd%r(nl(i)%aid,:), box, pf(i))
  enddo

! add pair forces to atom force
  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0 .and. nl(i)%nab .ne. 0) then
    if(nl(i)%inab .ne. 0) then
     nl(i)%fi = 0.d0
     call sum_pair_forces(nl(i),pf(i)%fij(1:nl(i)%inab,:))
     do j = 1, nl(i)%np
      atd%f(nl(i)%aid(j),:) = atd%f(nl(i)%aid(j),:) + nl(i)%fi(j,:)
      enddo
    endif
   endif
  enddo

! calculate self-bonded and Fourier series part of the electrostatic energy
! and forces and their contributions to the atomic pressure tensor
  do i = 1, spd%npsp
   if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
    call ewald_fourier(ewd, spd, atd, md, box, pta, ptm, jqm)
    call ewald_self_bonded(ewd, atd, spd, pta)
    p%inst(p%i(47)) = ewd%ufour
    p%inst(p%i(49)) = ewd%usb
    p%inst(p%i(48)) = ewd%uself
    p%inst(p%i(50)) = (ewd%psb + ewd%pfour)/box%volume
    exit ! only need to do it once, for all charges
   endif
  enddo

! add dihedral forces and energies to atomic force and energy
! and add dihedral contribution to atomic pressure tensor
  if(da%dfarch .ne. 0) call add_dihedral_forces(atd,da,md,ca,p%inst(p%i(41)),pta)
! add bond stretching forces and energies to atomic force and energy
! and add bond stretching contribution to atomic pressure tensor
  if(bs%bstype .ne. 0) call add_bondstr_forces(atd,bs,md,p%inst(p%i(37)), &
  & p%inst(p%i(38)),pta)
  if(bb%bbtype .ne. 0) call add_bondbend_forces(atd,bb,md,p%inst(p%i(57)), &
  & p%inst(p%i(58)),pta)

! sum force on each molecule
  call calc_fm(md, atd%f)

  end subroutine calc_pe_forces
!***********************************************************************
  subroutine calc_fluxes
! compute molecular pressure tensor and molecular heat flux vector
!***********************************************************************
  integer :: i, j

! Kinetic part of pressure tensor and heat flux vector
  ptm = ptm + sum_psq_on_m(md%pcm,md%massinv)
  jqm = jqm + jqkinetic(md%pcm,md%eint,md%massinv)

! Calculate molecular cm separations for interacting sites in each neighbour list
! then configurational part of molecular pressure tensor for each interaction
  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0 .and. nl(i)%nab .ne. 0) then
    if(nl(i)%inab .ne. 0) then
     j = nl(i)%inab
     call calc_rcmij(md, pf(i)%rij(1:j,:), nl(i)%aid(nl(i)%ilist(1:j,1)), &
     & nl(i)%aid(nl(i)%ilist(1:j,2)))
     ptm = ptm - sum_rij_fij(md%rcmij, pf(i)%fij(1:j,:))
     jqm = jqm - jqpotential(md%rcmij, pf(i)%fij(1:j,:), &
         & atd%p(nl(i)%aid(nl(i)%ilist(1:j,1)),:), &
         & atd%p(nl(i)%aid(nl(i)%ilist(1:j,2)),:), &
         & spd%mass(spd%spid(nl(i)%aid(nl(i)%ilist(1:j,1)))), &
         & spd%mass(spd%spid(nl(i)%aid(nl(i)%ilist(1:j,2)))))
    endif
   endif
  enddo

  call destroy_rcmij(md)

  ptm = ptm/box%volume
  jqm = jqm/box%volume

! molecular pressure tensor
  do i = 1, 3
   do j = 1, 3
    p%inst(p%i(22)-1+(i-1)*3+j) = ptm(i,j)
   enddo
  enddo

  do i = 1, 3
! trace of molecular pressure tensor
   p%inst(p%i(21)) = p%inst(p%i(21)) + ptm(i,i)
! molecular diffusive mass flux
   do j = 1, md%nf-1
    p%inst(p%i(8)+(j-1)*3+i-1) = sum(md%pcm(md%mf(j):md%ml(j),i))/box%volume
   enddo
! molecular heat flux vector
   p%inst(p%i(9)+i-1) = jqm(i)
  enddo
! pressure = tr(P)/dimensionality
  p%inst(p%i(21)) = p%inst(p%i(21))/3

  end subroutine calc_fluxes
!***********************************************************************
  subroutine do_correctors
!***********************************************************************
  
  if(sys%td%thermotype .eq. 2) then
   call corrector(sys%td%zeta,ded%zder%d1,ded%zder%d2,ded%zder%d3,ded%zder%d4,zeqm,ded%tstep)
  endif

  if(sys%prd%barotype .ne. 0) then
   if(sys%prd%barotype .eq. 2) call corrector &
    (sys%prd%edot,ded%eder%d1,ded%eder%d2,ded%eder%d3,ded%eder%d4,eeqm,ded%tstep)
   call corrector(sys%prd%volume,ded%vder%d1,ded%vder%d2,ded%vder%d3,ded%vder%d4,veqm,ded%tstep)
! scale box vectors and recalculate volume
   box%boxvec = box%boxvec*(sys%prd%volume/box%volume)**(1.d0/3.d0)
   box%volume = volume(box%boxvec)
  endif

! call correctors for primary particle positions
  call corrector(atd%r,ded%rder%d1,ded%rder%d2, &
   ded%rder%d3,ded%rder%d4,reqm,ded%tstep,ca%is_basic,dr)

  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0) call increment_nlist_drtot(nl(i),dr(nl(i)%aid,:))
  enddo

! call correctors for primary particle momenta
  call corrector(atd%p,ded%pder%d1,ded%pder%d2, &
   ded%pder%d3,ded%pder%d4,peqm,ded%tstep,ca%is_basic)

! update positions and momenta for secondary particles
  if(ca%any_secondary) call update_secondary(atd,md,spd,ca,dr) 

  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0) call increment_nlist_drtot(nl(i),dr(nl(i)%aid,:))
  enddo

  end subroutine do_correctors
!***********************************************************************
  subroutine do_pbcs
! Apply periodic boundary conditions to put all molecular cm in the box
!***********************************************************************
  call calc_rcm(md, spd)

   call apply_pbcs(box, md%rcm, rsh)
   atd%r = atd%r - rsh(md%mid,:)
   if(ded%destype .eq. 1) &
   call gear5_pbc_correction(ded%rder, ded%tstep, rsh(md%mid,:), &
    sd, ca%is_basic)

  end subroutine do_pbcs
!***********************************************************************
  subroutine do_predictors
!***********************************************************************
! predictor for primary particles
  
  call predictor(atd%r,ded%rder%d1,ded%rder%d2, &
  & ded%rder%d3,ded%rder%d4,ca%is_basic,dr)

  call predictor(atd%p,ded%pder%d1,ded%pder%d2, &
  & ded%pder%d3,ded%pder%d4,ca%is_basic)

! update the displacements or relative displacements used for the
! neighbour list reformation criterion
  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0) call increment_nlist_drtot(nl(i),dr(nl(i)%aid,:))
  enddo

! update positions and momenta for secondary particles
  if(ca%any_secondary) then
   call update_secondary(atd, md, spd, ca, dr)
! update dr for secondary particles
   do i = 1, spd%npsp
    if(pf(i)%pftype .ne. 0) call increment_nlist_drtot(nl(i),dr(nl(i)%aid,:))
   enddo
  endif

  if(sys%td%thermotype .eq. 2) &
  & call predictor(sys%td%zeta,ded%zder%d1,ded%zder%d2,ded%zder%d3,ded%zder%d4)

  if(sys%prd%barotype .ne. 0) then
   if(sys%prd%barotype .eq. 2) call predictor(sys%prd%edot,ded%eder%d1, &
   & ded%eder%d2,ded%eder%d3,ded%eder%d4)
   call predictor(sys%prd%volume,ded%vder%d1,ded%vder%d2,ded%vder%d3,ded%vder%d4)
  endif

  end subroutine do_predictors
!***********************************************************************
  subroutine evolve_box
!***********************************************************************
! Evolve box vectors with flow and calculate new volume, inverse box vectors
! and k-vectors for Ewald summations.

  if(sd%flow_type .ge. 1) then
   call evolve_box_vectors(box%boxvec, box%boxvec0, sd, ded%tstep)
   box%volume = volume(box%boxvec)
   box%boxveci = matinv3d(transpose(box%boxvec))
   do i = 1, spd%npsp
    if(pf(i)%pftype .eq. 3 .or. pf(i)%pftype .eq. 4) then
     call init_ewald_kvec(box%boxveci, ewd)
     exit
    endif
   enddo
  endif

  end subroutine evolve_box
!***********************************************************************
  subroutine g5step
!***********************************************************************

  call init_forces_properties
  call do_predictors
  call evolve_box
  call calc_pe_forces
  call init_eqm
  call calc_constraint_forces
  call calc_props_2
  call calc_eqm
  call do_correctors

  end subroutine g5step
!***********************************************************************
  subroutine init_eqm
!***********************************************************************
! initialize equations of motion
  do i = 1, atd%n
   reqm(i,:) = atd%p(i,:)/spd%mass(spd%spid(i))   ! rdot equation
  enddo
  peqm = atd%f                                    ! pdot equation
  zeqm = 0.d0
  eeqm = 0.d0
  veqm = 0.d0

  end subroutine init_eqm
!***********************************************************************
  subroutine init_mdstep
!***********************************************************************
! Increment the number of timesteps
  call update_step(rc)
! Check whether to do accumulation of properties or time series this timestep
  call check_properties_flags(rc%nstep, p)

  end subroutine init_mdstep
!***********************************************************************
  subroutine init_forces_properties
!***********************************************************************
! zero instantaneous values of properties
  p%inst = 0.d0

! zero forces, internal energy pressure tensors and heat flux vector
  atd%f = 0.d0
  atd%eint = 0.d0
  pta = 0.d0
  ptm = 0.d0
  jqm = 0.d0
  alpha = 0.d0

  end subroutine init_forces_properties
!***********************************************************************
  subroutine rescale_cm_momenta
!***********************************************************************
  double precision :: sf

! Rescale momenta if required
   if(sys%td%ktsc .gt. 0) then
   if(mod(rc%nstep,sys%td%ktsc) .eq. 0) then
    call rescale_momenta(md%pcm, sys%td%tin, sys%td%ndfm, md%massinv,sf)
    do i = 1, md%nm
     do k = 1, md%nsf(md%fid(i))
      do j = 1, 3
       md%mol(i)%p(k,j) = md%mol(i)%p(k,j)*sf
      enddo
     enddo
    enddo
   endif
   endif

  end subroutine rescale_cm_momenta
!***********************************************************************
  subroutine rk4step(np, ndim)
!***********************************************************************
  integer :: np, ndim
  double precision, dimension(np,ndim) :: rold, pold
  double precision, dimension(np,ndim,4) :: rk4kr, rk4kp
  double precision, dimension(4) :: rk4kz, rk4ked, rk4kv
  double precision :: zold, edold, vold

   rold = atd%r
   pold = atd%p
   zold = sys%td%zeta
   edold = sys%prd%edot
   vold = sys%prd%volume

! Begin rk4k1
   call init_forces_properties
   call calc_pe_forces
   call init_eqm
   call calc_constraint_forces
   call calc_props_2
   call calc_eqm
   do i = 1, size(atd%r,2)
    where(ca%is_basic)
     rk4kr(:,i,1) = reqm(:,i)*ded%tstep
     rk4kp(:,i,1) = peqm(:,i)*ded%tstep
     atd%r(:,i) = rold(:,i) + rk4kr(:,i,1)*0.5d0
     atd%p(:,i) = pold(:,i) + rk4kp(:,i,1)*0.5d0
    endwhere
   enddo
! update positions and momenta for secondary particles
   if(ca%any_secondary) call update_secondary(atd, md, spd, ca)

  if(sys%td%thermotype .eq. 2) then
    rk4kz = zeqm*ded%tstep
    sys%td%zeta = zold + rk4kz(1)*0.5d0
  endif

  if(sys%prd%barotype .ne. 0) then
   if(sys%prd%barotype .eq. 2) then
    rk4ked(1) = eeqm*ded%tstep
    sys%prd%edot = edold + rk4ked(1)*0.5d0
   endif
   rk4kv(1) = veqm*ded%tstep
   sys%prd%volume = vold + rk4kv(1)*0.5d0
! scale box vectors and recalculate volume
   box%boxvec = box%boxvec*(sys%prd%volume/box%volume)**(1.d0/3.d0)
   box%volume = volume(box%boxvec)
  endif

! should evolve box by a half step - explicit time-dependence
   call evolve_box

! Begin rk4k2
   call init_forces_properties
   call calc_pe_forces
   call init_eqm
   call calc_constraint_forces
   call calc_props_2
   call calc_eqm
   do i = 1, size(atd%r,2)
    where(ca%is_basic)
     rk4kr(:,i,2) = reqm(:,i)*ded%tstep
     rk4kp(:,i,2) = peqm(:,i)*ded%tstep
     atd%r(:,i) = rold(:,i) + rk4kr(:,i,2)*0.5d0
     atd%p(:,i) = pold(:,i) + rk4kp(:,i,2)*0.5d0
    endwhere
   enddo
! update positions and momenta for secondary particles
   if(ca%any_secondary) call update_secondary(atd, md, spd, ca)

  if(sys%td%thermotype .eq. 2) then
    rk4kz(2) = zeqm*ded%tstep
    sys%td%zeta = zold + rk4kz(2)*0.5d0
  endif

  if(sys%prd%barotype .ne. 0) then
   if(sys%prd%barotype .eq. 2) then
    rk4ked(2) = eeqm*ded%tstep
    sys%prd%edot = edold + rk4ked(2)*0.5d0
   endif
   rk4kv(2) = veqm*ded%tstep
   sys%prd%volume = vold + rk4kv(2)*0.5d0
! scale box vectors and recalculate volume
   box%boxvec = box%boxvec*(sys%prd%volume/box%volume)**(1.d0/3.d0)
   box%volume = volume(box%boxvec)
  endif

! Begin rk4k3
   call init_forces_properties
   call calc_pe_forces
   call init_eqm
   call calc_constraint_forces
   call calc_props_2
   call calc_eqm
   do i = 1, size(atd%r,2)
   where(ca%is_basic)
    rk4kr(:,i,3) = reqm(:,i)*ded%tstep
    rk4kp(:,i,3) = peqm(:,i)*ded%tstep
    atd%r(:,i) = rold(:,i) + rk4kr(:,i,3)
    atd%p(:,i) = pold(:,i) + rk4kp(:,i,3)
   endwhere
   enddo
! update positions and momenta for secondary particles
  if(ca%any_secondary) call update_secondary(atd, md, spd, ca)

  if(sys%td%thermotype .eq. 2) then
    rk4kz(3) = zeqm*ded%tstep
    sys%td%zeta = zold + rk4kz(3)
  endif

  if(sys%prd%barotype .ne. 0) then
   if(sys%prd%barotype .eq. 2) then
    rk4ked(3) = eeqm*ded%tstep
    sys%prd%edot = edold + rk4ked(3)
   endif
   rk4kv(3) = veqm*ded%tstep
   sys%prd%volume = vold + rk4kv(3)
! scale box vectors and recalculate volume
   box%boxvec = box%boxvec*(sys%prd%volume/box%volume)**(1.d0/3.d0)
   box%volume = volume(box%boxvec)
  endif

! should evolve box by another half step
!   call evolve_box

! Begin rk4k4
   call init_forces_properties
   call calc_pe_forces
   call init_eqm
   call calc_constraint_forces
   call calc_props_2
   call calc_eqm
   do i = 1, size(atd%r,2)
    where(ca%is_basic)
     rk4kr(:,i,4) = reqm(:,i)*ded%tstep
     rk4kp(:,i,4) = peqm(:,i)*ded%tstep

     dr(:,i) = (rk4kr(:,i,1) + 2.d0*rk4kr(:,i,2) &
      + 2.d0*rk4kr(:,i,3) + rk4kr(:,i,4))/6.d0

     atd%r(:,i) = rold(:,i) + dr(:,i)
     atd%p(:,i) = pold(:,i) + (rk4kp(:,i,1) + 2.d0*rk4kp(:,i,2) &
      + 2.d0*rk4kp(:,i,3) + rk4kp(:,i,4))/6.d0
    endwhere
   enddo
! update positions and momenta for secondary particles
  if(ca%any_secondary) call update_secondary(atd, md, spd, ca, dr)

! need to update dr here for neighbour list after moving secondary particles
! update the displacements or relative displacements used for the
! neighbour list reformation criterion
  do i = 1, spd%npsp
   if(pf(i)%pftype .ne. 0) call increment_nlist_drtot(nl(i),dr(nl(i)%aid,:))
  enddo

  if(sys%td%thermotype .eq. 2) then
    rk4kz(4) = zeqm*ded%tstep
    sys%td%zeta = zold + (rk4kz(1) + 2.d0*rk4kz(2) &
          + 2.d0*rk4kz(3) + rk4kz(4))/6.d0
  endif

  if(sys%prd%barotype .ne. 0) then
   if(sys%prd%barotype .eq. 2) then
    rk4ked(1) = eeqm*ded%tstep
    sys%prd%edot = edold + (rk4ked(1) + 2.d0*rk4ked(2) &
          + 2.d0*rk4ked(3) + rk4ked(4))/6.d0
   endif
   rk4kv(1) = veqm*ded%tstep
   sys%prd%volume = vold + (rk4kv(1) + 2.d0*rk4kv(2) &
          + 2.d0*rk4kv(3) + rk4kv(4))/6.d0
! scale box vectors and recalculate volume
   box%boxvec = box%boxvec*(sys%prd%volume/box%volume)**(1.d0/3.d0)
   box%volume = volume(box%boxvec)
  endif

  end subroutine rk4step
!***********************************************************************
  function spec_dens(i)
! compute array of identifiers for species i
!***********************************************************************
  double precision, dimension(atd%n) :: spec_dens
  integer :: i
  where(spd%spid .eq. i)
   spec_dens = 1.d0
  elsewhere
   spec_dens = 0.d0
  endwhere
  end function spec_dens
!***********************************************************************
  end subroutine mdstep
!***********************************************************************
