!***********************************************************************
  module run_control
!***********************************************************************
!23456789012345678901234567890123456789012345678901234567890123456789012
  use simulation_box

  implicit none

  private

  public :: rd_run_control, rd_run_control_state, &
          & store_run_control_state, update_step,  &
          & wr_snapshot, zero_nstep

  public :: run_control_type

  type run_control_type
   double precision :: bf(3)
   double precision :: rsc
   integer :: runtype, &
            & nstep, steps_this_run, &
            & kstore, ksnapshot, kboxsc, kpz(3), num_ens_mem
   logical :: store_state_now, do_snapshot_now, do_branch_now
   character(len=64) :: ifile, ofile, efile ! input, output and error files
   character(len=64) :: sfile ! snapshot file name prefix
  end type run_control_type

!*Values of constants*
! runtype:    0=from crystal generated internally or from a config file
!             1=from stored state, but zero averages of all properties
!             2=continue from a stored state and stored property averages

  contains
!***********************************************************************
  subroutine rd_run_control(rc,header,iunit)
!***********************************************************************
! Reads run_control variables from input file.
! 
!***********************************************************************

  type(run_control_type) :: rc
  
  integer :: i, j, iunit
  
  character(len=*) :: header
  character(len=32) :: rdstring

  rdstring = ''
  i = 0
  open(iunit,file=rc%ifile,status='old',form='formatted')
  do j = 1, 1000
   read(iunit,*, iostat=i) rdstring
   if(rdstring .eq. header .or. i .ne. 0 .or. j .ge. 999) exit
  enddo

  if (rdstring .eq. header) then
   read(iunit,*) rc%steps_this_run
   read(iunit,*) rc%runtype
   read(iunit,*) rc%kstore
   read(iunit,*) rc%ksnapshot, rc%rsc
   read(iunit,*) rc%kboxsc, rc%bf(1), rc%bf(2), rc%bf(3)
   read(iunit,*) rc%kpz(1), rc%kpz(2), rc%kpz(3)
   read(iunit,*) rc%num_ens_mem
  close(iunit)
  else
   close(iunit)
   stop 'run_control: input data not found'
  endif

! initialize logical variables
  rc%store_state_now = .false.
  rc%do_snapshot_now = .false.
  rc%do_branch_now = .false.

! check for invalid input values

  end subroutine rd_run_control
!***********************************************************************
  subroutine rd_run_control_state(rc,sunit)
!***********************************************************************  
  type(run_control_type) :: rc
  integer :: sunit

  read(sunit) rc%nstep

  end subroutine rd_run_control_state
!***********************************************************************
  subroutine store_run_control_state(rc,sunit)
!***********************************************************************  
  type(run_control_type) :: rc
  integer :: sunit

  write(sunit) rc%nstep

  end subroutine store_run_control_state
!***********************************************************************
   subroutine update_step(rc)
!***********************************************************************
   type(run_control_type) :: rc

   rc%nstep = rc%nstep + 1

   if(rc%kstore .ne. 0) rc%store_state_now = mod(rc%nstep,rc%kstore) .eq. 0
   if(rc%ksnapshot .ne. 0) rc%do_snapshot_now = mod(rc%nstep,rc%ksnapshot) .eq. 0

  end subroutine update_step
!***********************************************************************
  subroutine wr_snapshot(ratom, boxvec, labels, rc, iunit)
! writes input data to the output file
!***********************************************************************
  type(run_control_type) :: rc
  double precision, dimension(:,:) :: boxvec
  double precision, dimension(:,:) :: ratom
  character(len=2), dimension(:) :: labels
  character, parameter :: tab=char(9)
  integer :: iunit, i, j

  open (iunit,file=rc%sfile,status='old',position='append')
  write(iunit,*) size(ratom,1)
  write(iunit,*) (real(boxvec(i,i)*rc%rsc), tab, i = 1, size(ratom,2))
  do i = 1, size(ratom,1)
   write(iunit,*) labels(i), (real(ratom(i,j)*rc%rsc), tab, j = 1, size(ratom,2))
  enddo
  close(iunit)
  
  end subroutine wr_snapshot
!***********************************************************************
   subroutine zero_nstep(rc)
!***********************************************************************
   type(run_control_type) :: rc

   rc%nstep = 0

  end subroutine zero_nstep
!***********************************************************************
  end module run_control
!***********************************************************************
