  module v_sub_library
!***********************************************************************
! Module containing customised mathematical subroutines.
!
!23456789012345678901234567890123456789012345678901234567890123456789012

  implicit none
 
  private
 
  public :: lusolv, xy_array_rotate, gssolv, gssolv_1, &
          & tridiag, set_spind, count_sp, gssolv_sp, gssolv_sp_1, &
          & tr_div_dim

  contains
!***********************************************************************
  subroutine gssolv_sp(a_sp, ind_sp, b, xprev, tol, its)
!***********************************************************************
  implicit none
  integer :: j, k, its, imax=1000, n, curx
  double precision, intent(in) :: tol
  double precision, dimension(:,:), intent(inout) :: b, xprev
  double precision, dimension(size(b,1), size(b,2)) :: xtemp, x
  double precision, dimension(:,:), intent(in) :: a_sp
  integer, dimension(:), intent(in) :: ind_sp

  xtemp = 0.0

  n = size(b,1)

  iters: do its = 1, imax
   x = 0.0
   j = n + 1

   do curx = 1, n
    do k = 1, ind_sp(curx)
     x(curx,:) = x(curx,:) - (xprev(ind_sp(j),:) * a_sp(j,:))
     j = j + 1
    end do

    x(curx,:) = (x(curx,:) + b(curx,:)) / a_sp(curx,:)
      
    xtemp(curx,:) = xprev(curx,:)
    xprev(curx,:) = x(curx,:)
   end do
   if (all(abs(xtemp - xprev) < tol)) exit iters
  end do iters

  b = xprev

  end subroutine gssolv_sp
!***********************************************************************
  subroutine gssolv_sp_1(a_sp, ind_sp, b, xprev, tol, mits)
!***********************************************************************
  implicit none
  integer :: i, j, k, its, imax=1000, n, curx, nm, mits
  double precision, intent(in) :: tol
  double precision, dimension(:,:), intent(inout) :: b, xprev
  double precision, dimension(size(b,1)) :: xtemp, x
  double precision, dimension(:,:), intent(in) :: a_sp
  integer, dimension(:), intent(in) :: ind_sp


  n = size(b,1)
  nm = size(b,2)
  mits = 0

  do i = 1, nm
   if(all(b(:,i) .eq. 0.0)) cycle
   xtemp = 0.0
   iters: do its = 1, imax
    x = 0.0
    j = n + 1

    do curx = 1, n
     do k = 1, ind_sp(curx)
      x(curx) = x(curx) - (xprev(ind_sp(j),i) * a_sp(j,i))
      j = j + 1
     end do

     x(curx) = (x(curx) + b(curx,i)) / a_sp(curx,i)
      
     xtemp(curx) = xprev(curx, i)
     xprev(curx, i) = x(curx)
    end do
    if (all(abs(xtemp(:) - xprev(:,i)) < tol)) exit iters
   end do iters
   mits = max(its, mits)
  end do

  b = xprev

  end subroutine gssolv_sp_1
!***********************************************************************
  function count_sp (arr)
!***********************************************************************
! Determines the size of the sparse matrix index vector
!***********************************************************************
  implicit none
  double precision, dimension(:,:) :: arr
  integer :: i, j, nc, count_sp

  nc = size(arr,1)

! First, allocate nc places for the information on the number of non-zero
! elements in that row
  count_sp = nc
! Then, count the number of non-zero off-diagonal elements
  do i = 1, nc
   do j = i+1, nc
    if (arr(i,j) .ne. 0) count_sp = count_sp + 1
    if (arr(j,i) .ne. 0) count_sp = count_sp + 1
   enddo
  enddo

  end function count_sp
!***********************************************************************
  subroutine set_spind (arr, ind)
!***********************************************************************
! Allocates and assigns the vector containing the sparse matrix
! indexing information for the square double precision array arr.
!***********************************************************************
  implicit none
  double precision, dimension(:,:) :: arr
  integer, dimension(:) :: ind
  integer :: i, j, nr, ir, odi

  nr = size(arr,1)  ! number of rows, assumed the same as number of columns
  odi = nr   ! offset for off diagonal index information
  do i = 1, nr
   ir = 0
   do j = 1, nr
    if(i .eq. j .or. arr(i,j) .eq. 0) cycle
    odi = odi + 1
    ind(odi) = j
    ir = ir + 1
   enddo
   ind(i) = ir
  enddo

  end subroutine set_spind
!***********************************************************************
  subroutine xy_array_rotate (xy, theta)
!***********************************************************************
! Finds the x and y vector components of an array of vectors that have
! been rotated by an angle theta in the x-y plane.
! Equivalent to rotating the coordinate system by -theta
!***********************************************************************
  implicit none
  double precision, dimension(:,:) :: xy
  double precision, dimension(size(xy,1),size(xy,2)) :: xytrans
  double precision, intent(in):: theta

  xytrans(:,1) = cos(theta)*xy(:,1) - sin(theta)*xy(:,2)
  xytrans(:,2) = sin(theta)*xy(:,1) + cos(theta)*xy(:,2)

  xy = xytrans

  end subroutine xy_array_rotate
!***********************************************************************
  subroutine xy_rotate (xy, theta, xytrans)
!***********************************************************************
! Finds the x and y vector components of a vector that has been
! rotated by an angle theta in the x-y plane.
! Equivalent to rotating the coordinate system by -theta
!***********************************************************************
  implicit none
  double precision, dimension(:), intent(in):: xy
  double precision, dimension(:), intent(out):: xytrans
  double precision :: theta

  xytrans(1) = cos(theta)*xy(1) - sin(theta)*xy(2)
  xytrans(2) = sin(theta)*xy(1) + cos(theta)*xy(2)

  end subroutine xy_rotate
!***********************************************************************
  subroutine lusolv (a, b, n, nm)
!***********************************************************************
! Linear equation solver
! Uses LU factorization subroutine and forward and back substitution
! Ref. "FORTRAN 77, D. M. Monro"
! Rewritten by P. Daivis to run on the CM-5
!***********************************************************************
  implicit none
  integer i,j,n,nm
  double precision, dimension(nm) :: sum
  double precision a(n,n,nm), b(n,nm)

! do LU factorization of A
  call lufact(a,n,nm)
! forward substitute for Z
  do j = 1, n
   sum = b(j,:)
   do i = 1, j-1
    sum = sum - a(j,i,:)*b(i,:)
   enddo
   b(j,:) = sum/a(j,j,:)
  enddo
! back substitute for X
  do j = n, 1, -1
   sum = b(j,:)
   do i = j+1, n
    sum = sum - a(j,i,:)*b(i,:)
   enddo
   b(j,:) = sum
  enddo      

  end subroutine lusolv
!***********************************************************************
      subroutine lufact (a, n, nm)
!***********************************************************************
! LU factorization subroutine
! Ref. "FORTRAN 77, D. M. Monro"
! Rewritten by P. Daivis to run on the CM-5
!***********************************************************************
      implicit none
      integer i,j,k,n,nm
      double precision, dimension(nm) :: piv, sum
      double precision a(n,n,nm)

! column 1 of L is already column 1 of A
! find row 1 of U

      piv = a(1,1,:)
      do j = 2, n
       a(1,j,:) = a(1,j,:)/piv
      enddo
! now do the rest
      do k = 2, n
! find column k of L
       do j = k, n
        sum = a(j,k,:)
        do i = 1, k-1
         sum = sum - a(j,i,:)*a(i,k,:)
        enddo
        a(j,k,:) = sum
       enddo
! find row k of U, except diagonal elements (U(k,k) = 1)
       piv = a(k,k,:)
       do j = k+1, n
        sum = a(k,j,:)
        do i = 1, k-1
         sum = sum - a(k,i,:)*a(i,j,:)
        enddo
        a(k,j,:) = sum/piv
       enddo
      enddo

      end subroutine lufact
!***********************************************************************
  subroutine gssolv (a, b, xprev, tol, n, nm, its)
!***********************************************************************
! Linear equation solver
! Uses Gauss-Seidel iterative method
! Written by Mark C
!***********************************************************************
  implicit none
  integer, intent(in) :: n, nm
  integer :: i, curx, its, imax=1000
  double precision, intent(in) :: tol
  double precision, dimension(n, nm) :: xtemp, x, b, xprev
  double precision, dimension(n,n,nm), intent(in) :: a

  xtemp(:,:) = 0.0

  iters: do its = 1, imax
   x(:,:) = 0.0
   do curx = 1, n
    do i = 1, n
     if (curx /= i) x(curx,:) = x(curx,:) - xprev(i,:) * a(curx, i, :)
    end do
    x(curx,:) = (x(curx,:) + b(curx,:)) / a(curx, curx, :)
    xtemp(curx, :) = xprev(curx, :)
    xprev(curx, :) = x(curx, :)
   end do
   if (all(abs(xtemp-xprev) < tol)) exit iters
  end do iters
  b = xprev

  end subroutine gssolv
!***********************************************************************
  subroutine gssolv_1 (a, b, xprev, tol, n, nm, mits)
!***********************************************************************
! Linear equation solver
! Uses Gauss-Seidel iterative method
! Written by Mark C - modified by PD
! - loop over molecules
!***********************************************************************
  implicit none
  integer, intent(in) :: n, nm
  integer :: i, j, curx, its, mits, imax=1000
  double precision, intent(in) :: tol
  double precision, dimension(n, nm) :: xtemp, x, b, xprev
  double precision, dimension(n,n,nm), intent(in) :: a

  mits = 0
  xtemp(:,:) = 0.0
  do j = 1, nm
  iters: do its = 1, imax
   x(:,:) = 0.0
   do curx = 1, n
    do i = 1, n
     if (curx /= i) x(curx,j) = x(curx,j) - xprev(i,j) * a(curx, i, j)
    end do
    x(curx,j) = (x(curx,j) + b(curx,j)) / a(curx, curx, j)
    xtemp(curx, j) = xprev(curx, j)
    xprev(curx, j) = x(curx, j)
   end do
   if (all(abs(xtemp(:,j)-xprev(:,j)) < tol)) exit iters
  end do iters
  b(:,j) = xprev(:,j)
  mits = max(its, mits)
  enddo

  end subroutine gssolv_1
!***********************************************************************
  subroutine tridiag (a, b, c, r, x, n, nm)
!***********************************************************************
!***********************************************************************
! Linear equation solver: LU decomposition for tridiagonal matrices
! From Numerical recipes in Fortran - sec. 2.6
! The tridiagonal matrix is passed in as three vectors, a, b, c
! Modified by P. Daivis to solve nm sets simultaneously.
!***********************************************************************
  implicit none
  integer, intent(in) :: n, nm
  integer :: j
  double precision, dimension(n, nm) :: a, b, c, r, x, tmp1
  double precision, dimension(nm) :: t1

  if(any(b(1,:) .eq. 0.)) stop 'zero value found in tridiag'
  t1 = b(1,:)
  x(1,:) = r(1,:)/t1

! Decomposition and forward substitution
  do j = 2, n
   tmp1(j,:) = c(j-1,:)/t1
   t1 = b(j,:) - a(j,:)*tmp1(j,:)
   if(any(t1 .eq. 0.)) stop 'tridiag linear equation solver failed'
   x(j,:) = (r(j,:) - a(j,:)*x(j-1,:))/t1
  enddo

! Backsubstitution
  do j = n-1, 1, -1
   x(j,:) = x(j,:) - tmp1(j+1,:)*x(j+1,:)
  enddo

  end subroutine tridiag
!***********************************************************************
  function tr_div_dim(tensor)
! Finds the trace of a second rank tensor and divides by the
! dimensionality.
!***********************************************************************
  implicit none
  double precision, dimension(:,:), intent(in):: tensor
  double precision :: tr_div_dim
  integer :: i, ndim
  
  ndim = size(tensor,1)
  tr_div_dim = 0.d0
  do i = 1, ndim
   tr_div_dim = tr_div_dim + tensor(i,i)
  enddo
  tr_div_dim = tr_div_dim/ndim

  end function tr_div_dim
!***********************************************************************
!***********************************************************************
  end module v_sub_library
